'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _babelTypes = require('babel-types');

var t = _interopRequireWildcard(_babelTypes);

var _identifiers = require('./identifiers');

var identifiers = _interopRequireWildcard(_identifiers);

var _predicates = require('./predicates');

var predicates = _interopRequireWildcard(_predicates);

var _templates = require('./templates');

var templates = _interopRequireWildcard(_templates);

var _utils = require('./utils');

var utils = _interopRequireWildcard(_utils);

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

var replacePrivateMethodsCallsByExternalizedOnes = function replacePrivateMethodsCallsByExternalizedOnes(classBodyPath, privateMethodsNames) {
  var className = utils.getClassName(classBodyPath);

  classBodyPath.traverse({
    CallExpression: function CallExpression(callExpressionPath) {
      if (predicates.isPrivateMethodCall(callExpressionPath.node, privateMethodsNames)) {
        var methodName = callExpressionPath.node.callee.property.name;

        callExpressionPath.replaceWith(t.callExpression(t.memberExpression(identifiers.getPrivateMethodsMap(className), t.identifier(methodName)), [t.thisExpression()].concat(_toConsumableArray(callExpressionPath.node.arguments))));
      }
    }
  });
};

var checkIfHasPrivateAttributesAndReplaceThemByWeakMapAccess = function checkIfHasPrivateAttributesAndReplaceThemByWeakMapAccess(classBodyPath, privateMethodsNames) {
  var className = utils.getClassName(classBodyPath);
  var hasPrivateAttributes = false;

  classBodyPath.traverse({
    MemberExpression: function MemberExpression(memberExpressionPath) {
      if (predicates.isPrivateAttributeAccess(memberExpressionPath.node, privateMethodsNames)) {
        hasPrivateAttributes = true;
        var attributeName = memberExpressionPath.node.property.name;
        memberExpressionPath.replaceWith(templates.makeWeakMapRead({
          WEAK_MAP: identifiers.getWeakMap(className),
          PROPERTY: t.identifier(attributeName)
        }));
      }
    }
  });

  return hasPrivateAttributes;
};

var declareAndInitializeWeakMap = function declareAndInitializeWeakMap(classBodyPath) {
  var className = utils.getClassName(classBodyPath);
  var weakMapIdentifier = identifiers.getWeakMap(className);
  utils.getExternalizedPrivatePropertiesDeclarationSiblingPath(classBodyPath).insertBefore(templates.makeWeakMapDeclaration({ WEAK_MAP: weakMapIdentifier }));

  var weakMapInitialization = templates.makeWeakMapInitialization({ WEAK_MAP: weakMapIdentifier });
  var constructorPath = utils.getConstructorPath(classBodyPath);
  if (constructorPath) {
    var firstStatementPath = constructorPath.get('body.body.0');
    if (utils.isSuperConstructionCall(firstStatementPath)) {
      firstStatementPath.insertAfter(weakMapInitialization);
    } else {
      firstStatementPath.insertBefore(weakMapInitialization);
    }
  } else {
    var firstMethodPath = classBodyPath.get('body').find(function (path) {
      return path.isClassMethod();
    });
    firstMethodPath.insertBefore(t.classMethod('constructor', t.identifier('constructor'), [], t.blockStatement([weakMapInitialization])));
  }
};

var declarePrivateMethodsMap = function declarePrivateMethodsMap(classBodyPath, privateMethodsNames) {
  var className = utils.getClassName(classBodyPath);

  return utils.getExternalizedPrivatePropertiesDeclarationSiblingPath(classBodyPath, 1).insertBefore(t.variableDeclaration('const', [t.variableDeclarator(identifiers.getPrivateMethodsMap(className), t.objectExpression(classBodyPath.node.body.filter(function (node) {
    return privateMethodsNames.indexOf(node.key.name) !== -1;
  }).map(function (node) {
    return t.objectMethod('method', node.key, [identifiers.getClassInstance(className)].concat(_toConsumableArray(node.params)), node.body);
  })))]))[0];
};

var replaceThisExpressionByClassInstanceVariable = function replaceThisExpressionByClassInstanceVariable(path, className) {
  path.traverse({
    ThisExpression: function ThisExpression(thisExpressionPath) {
      thisExpressionPath.replaceWith(identifiers.getClassInstance(className));
    }
  });
};

var removePrivateMethodsFromClassBody = function removePrivateMethodsFromClassBody(classBodyPath, privateMethodsNames) {
  classBodyPath.traverse({
    ClassMethod: function ClassMethod(classMethodPath) {
      if (privateMethodsNames.indexOf(classMethodPath.node.key.name) !== -1) {
        classMethodPath.remove();
      }
    }
  });
};

var externalizePivateMethods = function externalizePivateMethods(classBodyPath, privateMethodsNames) {
  var privateMethodsMapPath = declarePrivateMethodsMap(classBodyPath, privateMethodsNames);
  replaceThisExpressionByClassInstanceVariable(privateMethodsMapPath, utils.getClassName(classBodyPath));
  removePrivateMethodsFromClassBody(classBodyPath, privateMethodsNames);
};

var makePropertiesStartingWithAnUnderscoreTrulyPrivate = function makePropertiesStartingWithAnUnderscoreTrulyPrivate(classBodyPath) {
  var privateMethodsNames = utils.getPrivateMethodsNames(classBodyPath);

  if (privateMethodsNames.length) {
    replacePrivateMethodsCallsByExternalizedOnes(classBodyPath, privateMethodsNames);
  }

  if (checkIfHasPrivateAttributesAndReplaceThemByWeakMapAccess(classBodyPath, privateMethodsNames)) {
    declareAndInitializeWeakMap(classBodyPath);
  }

  if (privateMethodsNames.length) {
    externalizePivateMethods(classBodyPath, privateMethodsNames);
  }
};

var traverseWithClassBodyVisitor = function traverseWithClassBodyVisitor(path) {
  path.traverse({
    ClassBody: function ClassBody(classBodyPath) {
      makePropertiesStartingWithAnUnderscoreTrulyPrivate(classBodyPath);
    }
  });
};

exports.default = function () {
  return {
    // Using classDeclaration and classExpression instead of only ClassBody to be called before transfrom-es2015-classes
    visitor: {
      ClassDeclaration: function ClassDeclaration(path) {
        traverseWithClassBodyVisitor(path);
      },
      ClassExpression: function ClassExpression(path) {
        traverseWithClassBodyVisitor(path);
      }
    }
  };
};